#ifndef __XRDDPMXMIQUEUEANDREQ_H_
#define __XRDDPMXMIQUEUEANDREQ_H_
/******************************************************************************/
/*                                                                            */
/*               X r d D P M X m i Q u e u e A n d R e q . h h                */
/*                                                                            */
/* DPM handling and various by David Smith, CERN                              */            
/* Other parts:                                                               */
/*                                                                            */
/* (c) 2006 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/******************************************************************************/

#include "XrdOlb/XrdOlbXmiReq.hh"
#include "XrdSys/XrdSysError.hh"
#include "XrdOuc/XrdOucName2Name.hh"
#include "XrdSys/XrdSysPthread.hh"
#include "XrdSys/XrdSysSemWait.hh"

#include "osdep.h"
#include "dpns_api.h"

struct XrdOlbXmiEnv;
class  XrdOucTrace;

class XrdDPMReq;
class XrdDPMXmi;

class XrdDPMQueue
{
       friend class XrdDPMReq;
public:
                              XrdDPMQueue();
                             ~XrdDPMQueue() {}
       XrdDPMReq             *Alloc(const XrdOlbXmiReq::ReqType rtype, XrdOlbReq *reqP, const char *Path, void *pd);
       void                   Set(XrdDPMXmi *xmi, XrdOlbXmiEnv *Env);
       int                    MakeOnePoll(int &nqueued);
       int                    Wait4Q(const int &prevnum, const int &max);

private:
       int                    do_poll_switch(XrdDPMReq *p, const bool &destruct);
       void                   Remove(const XrdDPMReq *toDel);

       XrdDPMXmi             *myXmi;
       XrdSysMutex            queueMutex;
       XrdSysSemWait          queueSemWait;
       XrdDPMReq             *nextFree;
       XrdDPMReq             *firstQueue;
       XrdDPMReq             *lastQueue;
       int                    numFree;
       int                    numinQ;
static const int              maxFree     =   100;
static const int              retryTime   =   60;
static const int              maxqueueage =  3600; //    Max time for a request not to progress
       XrdSysError           *eDest;               // -> Error message handler
       XrdOucTrace           *Trace;               // -> Trace handler
       XrdOucName2Name       *N2N;                 // -> lfn mapper
       char                   isWaiting ;
};

class XrdDPMReq
{
       friend class XrdDPMQueue;
public:
                              XrdDPMReq() {}
                             ~XrdDPMReq() {}
       void                   init(XrdOlbReq *reqP, XrdOlbXmiReq::ReqType rtype, XrdDPMQueue *queue, void *pd);
const  char                  *Path() {return thePath;}
       void                   Queue();
       XrdOlbReq             *Request() {return olbReq;}
       void                   notifyActionTaken(const bool &madeProgress);
       void                  *getPd() {return pd;}

       XrdDPMReq             *Next;
       XrdDPMReq             *Prev;

enum   reqObjState {
          never_called, first_action, waiting_action, waiting_fast_action
       };

struct sopts_s {
          bool have_stoken;
          char stoken[CA_MAXDPMTOKENLEN+1];
          bool have_ftype;
          char ftype;
          bool have_lifetime;
          time_t lifetime;
          bool have_reqsize;
          u_signed64 reqsize;
       } sopts;

private:
       void                   Recycle();
       XrdDPMQueue           *myQueue;
       XrdOlbXmiReq::ReqType olbXmiRtype;
       XrdOlbReq             *olbReq;
static const unsigned int     PathSize = CA_MAXPATHLEN+1;
       char                   thePath[PathSize];
       reqObjState            state;
       time_t                 pauseTime;
       time_t                 lastProgressMade;
       time_t                 lastPollStarted;
       void                   *pd; // private data meaningful to the processing routine
};
#endif
